# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
This module handles the conversion of various VOTABLE datatypes
to/from TABLEDATA_ and BINARY_ formats.
"""

# STDLIB
import codecs
import re
from struct import unpack as struct_unpack
from struct import pack as struct_pack

# THIRD-PARTY
import numpy as np

# LOCAL
from xmlutil import xml_escape

array_splitter = re.compile(r"\s+|(?:\s*,\s*)")
"""
A regex to handle splitting values on either whitespace or commas.

SPEC: Usage of commas is not actually allowed by the spec, but many
files in the wild use them.
"""

_zero_int = '\0\0\0\0'

class Converter(object):
    """
    The base class for all converters.  Each subclass handles
    converting a specific VOTABLE data type to/from the TABLEDATA_ and
    BINARY_ on-disk representations.
    """
    def __init__(self, field, config={}):
        """
        *field* is a :class:`vo.table.Field` object describing the
         datatype.
        """
        pass

    @staticmethod
    def _parse_length(read):
        return struct_unpack(">I", read(4))[0]

    @staticmethod
    def _write_length(length):
        return struct_pack(">I", length)

    def parse(self, value):
        """
        Convert the string *value* from the TABLEDATA_ format into an
        object with the correct native in-memory datatype.
        """
        raise NotImplementedError(
            "This datatype must implement a 'parse' method.")

    def output(self, value):
        """
        Convert the object *value* in the native in-memory datatype to
        a string suitable for serializing in the TABLEDATA_ format.
        """
        raise NotImplementedError(
            "This datatype must implement a 'output' method.")

    def binparse(self, read):
        """
        Reads some number of bytes from the BINARY_ format
        representation by calling the function *read*, and returns the
        native in-memory object representation for the datatype
        handled by *self*.
        """
        raise NotImplementedError(
            "This datatype must implement a 'binparse' method.")

    def binoutput(self, value):
        """
        Convert the object *value* in the native in-memory datatype to
        a string of bytes suitable for serialization in the BINARY_
        format.
        """
        raise NotImplementedError(
            "This datatype must implement a 'binoutput' method.")


class Char(Converter):
    """
    Handles the char datatype. (8-bit unsigned characters)
    """
    default = ''

    def __init__(self, field, config={}):
        Converter.__init__(self, field, config)

        if field.arraysize is None or field.arraysize == '*':
            self.format = 'O'
            self.binparse = self._binparse_var
            self.binoutput = self._binoutput_var
        else:
            if field.arraysize.endswith('*'):
                field.arraysize = field.arraysize[:-1]
            try:
                self.arraysize = int(field.arraysize)
            except ValueError:
                raise ValueError(
                    "invalid size specifier for a char field")
            self.format = 'S%d' % self.arraysize
            self.binparse = self._binparse_fixed
            self.binoutput = self._binoutput_fixed
            self._struct_format = ">%ds" % self.arraysize

        if config.get('pedantic'):
            self.parse = self._ascii_parse
        else:
            self.parse = str

    @staticmethod
    def _ascii_parse(value):
        return value.encode('ascii')

    @staticmethod
    def output(value):
        return xml_escape(value).encode('ascii')

    def _binparse_var(self, read):
        length = self._parse_length(read)
        return read(length).decode('ascii')

    def _binparse_fixed(self, read):
        s = struct_unpack(self._struct_format, read(self.arraysize))[0]
        end = s.find('\0')
        if end != -1:
            return s[:end]
        return s

    def _binoutput_var(self, value):
        if value is None or value == '':
            return _zero_int
        encoded = value.encode('ascii')
        return self._write_length(len(encoded)) + encoded

    def _binoutput_fixed(self, value):
        return struct_pack(self._struct_format, value.encode('ascii'))


class UnicodeChar(Converter):
    """
    Handles the unicodeChar data type. UTF-16-BE.
    """
    default = u''

    def __init__(self, field, config={}):
        Converter.__init__(self, field, config)

        if field.arraysize is None or field.arraysize == '*':
            self.format = 'O'
            self.binparse = self._binparse_var
            self.binoutput = self._binoutput_var
        else:
            try:
                self.arraysize = int(field.arraysize)
            except ValueError:
                raise ValueError(
                    "char fields can not be multidimensional")
            self.format = 'U%d' % self.arraysize
            self.binparse = self._binparse_fixed
            self.binoutput = self._binoutput_fixed
            self._struct_format = ">%ds" % (self.arraysize * 2)

    @staticmethod
    def parse(value):
        return unicode(value, 'utf-8')

    @staticmethod
    def output(value):
        return xml_escape(value).encode('utf-8')

    def _binparse_var(self, read):
        length = self._parse_length(read)
        return read(length * 2).decode('utf_16_be')

    def _binparse_fixed(self, read):
        s = struct_unpack(self._struct_format, read(self.arraysize * 2))[0]
        s = s.decode('utf_16_be')
        end = s.find('\0')
        if end != -1:
            return s[:end]
        return s

    def _binoutput_var(self, value):
        if value is None or value == '':
            return _zero_int
        encoded = value.encode('utf_16_be')
        return self._write_length(len(encoded) / 2) + encoded

    def _binoutput_fixed(self, value):
        return struct_pack(self._struct_format, value.encode('utf_16_be'))


class VarArray(Converter):
    """
    Handles variable lengths arrays (i.e. where *arraysize* is '*').
    """
    format = 'O'
    default = []

    def __init__(self, field, base, arraysize, config={}):
        Converter.__init__(self, field, config)

        self._base = base

    def output(self, value):
        output = self._base.output
        result = [output(x) for x in value]
        return ' '.join(result)

    def binparse(self, read):
        length = self._parse_length(read)

        result = []
        binparse = self._base.binparse
        for i in xrange(length):
            result.append(binparse(read))
        return np.array(result)

    def binoutput(self, value):
        if value is None or len(value) == 0:
            return _zero_int

        length = len(value)
        result = [self._write_length(length)]
        binoutput = self._base.binoutput
        for x in value:
            result.append(binoutput(x))
        return ''.join(result)


class ArrayVarArray(VarArray):
    """
    Handles an array of variable-length arrays, i.e. where *arraysize*
    ends in '*'.
    """
    def parse(self, value):
        if value.strip() == '':
            parts = []
        parts = array_splitter.split(value)

        items = self._base._items
        parse_parts = self._base.parse_parts
        if len(parts) % items != 0:
            raise ValueError("Incorrect number of elements in array")
        result = []
        for i in xrange(0, len(parts), items):
            result.append(parse_parts(parts[i:i+items]))
        return np.array(result)


class ScalarVarArray(VarArray):
    def parse(self, value):
        if value.strip() == '':
            parts = []
        parts = array_splitter.split(value)

        parse = self._base.parse
        return np.array([parse(x) for x in parts], dtype=self._base.format)


class NumericArray(Converter):
    vararray_type = ArrayVarArray

    def __init__(self, field, base, arraysize, config={}):
        Converter.__init__(self, field, config)

        self._base = base
        self._arraysize = arraysize
        self.format = "%s%s" % (tuple(arraysize), base.format)

        self._items = 1
        for dim in arraysize:
            self._items *= dim

        self._memsize = np.dtype(self.format).itemsize
        self._bigendian_format = '>' + self.format

        self.default = (
            np.ones(arraysize, dtype=self._base.format) *
            self._base.default)

    def parse(self, value):
        if value.strip() == '':
            parts = []
        parts = array_splitter.split(value)
        if len(parts) != self._items:
            raise ValueError("Incorrect number of elements in array")
        return self.parse_parts(parts)

    def parse_parts(self, parts):
        base_parse = self._base.parse
        result = np.array([base_parse(x) for x in parts],
                          dtype=self._base.format)
        return result.reshape(self._arraysize)

    def output(self, value):
        base_output = self._base.output
        return ' '.join(base_output(x) for x in value.flat)

    def binparse(self, read):
        result = np.fromstring(read(self._memsize),
                               dtype=self._bigendian_format)[0]
        return result

    def binoutput(self, value):
        big_endian = np.array(value, dtype=value.dtype.newbyteorder('>'))
        return big_endian.tostring()


class Numeric(Converter):
    """
    The base class for all numeric data types.
    """
    array_type = NumericArray
    vararray_type = ScalarVarArray

    def __init__(self, field, config={}):
        Converter.__init__(self, field, config)

        self._memsize = np.dtype(self.format).itemsize
        self._bigendian_format = '>' + self.format

    def binparse(self, read):
        result = np.fromstring(read(self._memsize),
                               dtype=self._bigendian_format)
        return result[0]

    @staticmethod
    def binoutput(value):
        return value.newbyteorder('>').tostring()


class FloatingPoint(Numeric):
    """
    The base class for floating-point datatypes.
    """
    default = np.nan

    def __init__(self, field, config={}):
        Numeric.__init__(self, field, config)

        precision = field.precision
        if precision is None:
            self._output_format = '%g'
        elif precision.startswith("E"):
            self._output_format = "%%.%dE" % (int(precision[1:]))
        elif precision.startswith("F"):
            self._output_format = "%%.%dg" % (int(precision[1:]))
        else:
            self._output_format = "%%.%dg" % (int(precision))

    parse = staticmethod(float)

    def output(self, value):
        if np.isfinite(value):
            return self._output_format % value
        elif np.isnan(value):
            return 'NaN'
        elif np.isposinf(value):
            return '+InF'
        elif np.isneginf(value):
            return '-InF'
        raise ValueError("Invalid floating point value")


class Double(FloatingPoint):
    """
    Handles the double datatype.  Double-precision IEEE
    floating-point.
    """
    format = 'f8'


class Float(FloatingPoint):
    """
    Handles the float datatype.  Single-precision IEEE floating-point.
    """
    format = 'f4'


class Integer(Numeric):
    """
    The base class for all the integral datatypes.
    """
    default = 0

    def __init__(self, field, config={}):
        Numeric.__init__(self, field, config)

        if config.get('support_hexadecimal'):
            self.parse = self._hex_parse
        else:
            self.parse = int

    @staticmethod
    def _hex_parse(value):
        try:
            return int(value)
        except ValueError:
            if value.startswith('0x'):
                return int(value, 16)
        # May still raise if hex doesn't work, but that's
        # what we want...

    output = staticmethod(str)


class UnsignedByte(Integer):
    """
    Handles the unsignedByte datatype.  Unsigned 8-bit integer.
    """
    format = 'u1'


class Short(Integer):
    """
    Handles the short datatype.  Signed 16-bit integer.
    """
    format = 'i2'


class Int(Integer):
    """
    Handles the int datatype.  Signed 32-bit integer.
    """
    format = 'i4'


class Long(Integer):
    """
    Handles the long datatype.  Signed 64-bit integer.
    """
    format = 'i8'


class ComplexArrayVarArray(VarArray):
    """
    Handles an array of variable-length arrays of complex numbers.
    """
    def __init__(self, field, base, arraysize, config={}):
        VarArray.__init__(self, field, base, arraysize, config)

    def parse(self, value):
        if value.strip() == '':
            parts = []
        parts = array_splitter.split(value)

        items = self._base._items
        parse_parts = self._base.parse_parts
        if len(parts) % items != 0:
            raise ValueError("Incorrect number of elements in array")
        result = []
        for i in xrange(0, len(parts), items):
            result.append(parse_parts(parts[i:i+items]))
        return np.array(result)


class ComplexVarArray(VarArray):
    """
    Handles a variable-length array of complex numbers.
    """
    def __init__(self, field, base, arraysize, config={}):
        VarArray.__init__(self, field, base, arraysize, config)

    def parse(self, value):
        if value.strip() == '':
            parts = []
        parts = array_splitter.split(value)

        parse = self._base.parse_parts
        result = []
        for i in xrange(0, len(parts), 2):
            result.append(parse(parts[i:i+1]))
        return np.array(result, dtype=self._base.format)


class ComplexArray(NumericArray):
    """
    Handles a fixed-size array of complex numbers.
    """
    vararray_type = ComplexArrayVarArray

    def __init__(self, field, base, arraysize, config={}):
        NumericArray.__init__(self, field, base, arraysize, config)
        self._items *= 2

    def parse(self, value):
        if value.strip() == '':
            parts = []
        parts = array_splitter.split(value)
        return self.parse_parts(parts)

    def parse_parts(self, parts):
        if len(parts) != self._items:
            raise ValueError("Incorrect number of elements in array")
        base_parse = self._base.parse_parts
        result = []
        for i in xrange(0, self._items, 2):
            result.append(base_parse(parts[i:i+1]))
        result = np.array(result, dtype=self._base.format)
        return result.reshape(self._arraysize)


class Complex(FloatingPoint):
    """
    The base class for complex numbers.
    """
    array_type = ComplexArray
    vararray_type = ComplexVarArray
    default = np.nan

    def __init__(self, field, config={}):
        FloatingPoint.__init__(self, field, config)

        self._output_format = self._output_format + " " + self._output_format

    def parse(self, value):
        parts = [float(x) for x in array_splitter.split(value)]
        if len(parts) != 2:
            raise ValueError("'%s' does not parse as a complex number" % value)
        return self.parse_parts(parts)

    def parse_parts(self, parts):
        return complex(*parts)

    def output(self, value):
        return self._output_format % (value.real, value.imag)


class FloatComplex(Complex):
    """
    Handle floatComplex datatype.  Pair of single-precision IEEE
    floating-point numbers.
    """
    format = 'c8'


class DoubleComplex(Complex):
    """
    Handle doubleComplex datatype.  Pair of double-precision IEEE
    floating-point numbers.
    """
    format = 'c16'


class BitArray(NumericArray):
    """
    Handles an array of bits.
    """
    vararray_type = ArrayVarArray

    def __init__(self, field, base, arraysize, config={}):
        NumericArray.__init__(self, field, base, arraysize, config)

        self._bytes = ((self._items - 1) // 8) + 1

    def binparse(self, read):
        data = read(self._bytes)
        results = []
        for byte in data:
            byte = ord(byte)
            for bit_no in range(7, -1, -1):
                bit = byte & (1 << bit_no)
                bit = (bit != 0)
                results.append(bit)
                if len(results) == self._items:
                    break
            if len(results) == self._items:
                break

        result = np.array(results, dtype='b1')
        return result.reshape(self._arraysize)

    def binoutput(self, value):
        value = value.flat
        length = len(value)
        bit_no = 7
        byte = 0
        bytes = []
        for v in value:
            if v:
                byte |= 1 << bit_no
            if bit_no == 0:
                bytes.append(byte)
                bit_no = 7
                byte = 0
            else:
                bit_no -= 1
        bytes.append(byte)
        return struct_pack("%sB" % len(bytes), *bytes)


class Bit(Converter):
    """
    Handles the bit datatype.
    """
    format = 'b1'
    array_type = BitArray
    vararray_type = ScalarVarArray
    default = False

    def __init__(self, field, config={}):
        Converter.__init__(self, field, config)

    def parse(self, value):
        mapping = {'1': True, '0': False}
        try:
            return mapping[value]
        except KeyError:
            raise ValueError("Invalid bit value")

    def output(self, value):
        if value:
            return '1'
        else:
            return '0'

    def binparse(self, read):
        data = read(1)
        return (ord(data) & 0x8) != 0

    def binoutput(self, value):
        if value:
            return chr(0x8)
        return chr(0)


class BooleanArray(NumericArray):
    """
    Handles an array of boolean values.
    """
    vararray_type = VarArray

    def __init__(self, field, base, arraysize, config={}):
        NumericArray.__init__(self, field, base, arraysize, config)

    def binparse(self, read):
        data = read(self._items)
        binparse = self._base.binparse_value
        result = np.array([binparse(char) for char in data], dtype='b1')
        return result.reshape(self._arraysize)

    def binoutput(self, value):
        binoutput = self._base.binoutput
        result = [binoutput(x) for x in value.flat]
        return ''.join(result)


class Boolean(Converter):
    """
    Handles the boolean datatype.
    """
    format = 'b1'
    array_type = BooleanArray
    vararray_type = VarArray
    default = False

    def __init__(self, field, config={}):
        Converter.__init__(self, field, config)

    def parse(self, value):
        mapping = {'TRUE' : True,
                   'FALSE': False}
        try:
            return mapping[value.upper()]
        except KeyError:
            raise ValueError("Invalid boolean value")

    def output(self, value):
        if value:
            return 'TRUE'
        return 'FALSE'

    def binparse(self, read):
        value = read(1)
        return self.binparse_value(value)

    def binparse_value(self, value):
        mapping = {'T': True,  't': True,  '1': True,
                   'F': False, 'f': False, '0': False,
                   '\0': None, ' ': None,  '?': None}
        try:
            return mapping[value]
        except KeyError:
            raise ValueError("Invalid boolean value")

    def binoutput(self, value):
        if value:
            return 'T'
        return 'F'


converter_mapping = {
    'double'        : Double,
    'float'         : Float,
    'bit'           : Bit,
    'boolean'       : Boolean,
    'unsignedByte'  : UnsignedByte,
    'short'         : Short,
    'int'           : Int,
    'long'          : Long,
    'floatComplex'  : FloatComplex,
    'doubleComplex' : DoubleComplex,
    'char'          : Char,
    'unicodeChar'   : UnicodeChar }


def get_converter(field, config):
    """
    Given a :class:`~vo.tree.Field` object *field*, return an
    appropriate converter class to handle the specified datatype.
    """
    if field.datatype not in converter_mapping:
        raise ValueError("Unknown datatype '%s'" % field.datatype)

    cls = converter_mapping[field.datatype]
    converter = cls(field, config)

    arraysize = field.arraysize

    # With numeric datatypes, special things need to happen for
    # arrays.
    if (field.datatype not in ('char', 'unicodeChar') and
        arraysize is not None):
        if arraysize[-1] == '*':
            arraysize = arraysize[:-1]
            last_x = arraysize.rfind('x')
            if last_x == -1:
                arraysize = ''
            else:
                arraysize = arraysize[:last_x]
            fixed = False
        else:
            fixed = True

        if arraysize != '':
            arraysize = [int(x) for x in arraysize.split("x")]
            arraysize.reverse()
        else:
            arraysize = []

        if arraysize != []:
            converter = converter.array_type(
                field, converter, arraysize, config)

        if not fixed:
            converter = converter.vararray_type(
                field, converter, arraysize, config)

    return converter
