# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
This file contains a contains the high-level functions to read a
VOTable file.
"""

# STDLIB
import sys

# LOCAL
from voexceptions import *
import tree
import xmlutil

def parse(source, columns=None, invalid='exception', pedantic=True,
          support_hexadecimal=False, chunk_size=tree.DEFAULT_CHUNK_SIZE):
    """
    Parses a VOTABLE xml file (or file-like object), and returns a
    :class:`~vo.tree.VOTable` object, with a nested list of
    :class:`~vo.tree.Resource` instances and :class:`~vo.tree.Table`
    instances.

    *source* may be a filename or a readable file-like object.

    If the *columns* parameter is specified, it should be a list of
    field names to include in the output.  The default is to include
    all fields.

    The *invalid* parameter may be one of the following values:

      - 'exception': throw an exception when an invalid value is
        encountered (default)

      - 'mask': mask out invalid values

    When *pedantic* is True, raise an error when the file violates the
    spec, otherwise issue a warning.  Warnings may be controlled using
    the standard Python mechanisms.  See the :mod:`warnings` module in
    the Python standard library for more information.

    When *support_hexadecimal* is True, integer fields may contain
    hexadecimal numbers.  When False, this part of the spec is
    unsupported for efficiency reasons.

    *chunk_size* is the number of rows to read before converting to an
    array.  Higher numbers are likely to be faster, but will consume
    more memory.
    """
    invalid = invalid.lower()
    assert invalid in ('exception', 'mask')

    config = {'columns':  columns,
              'invalid':  invalid,
              'pedantic': pedantic,
              'support_hexadecimal': support_hexadecimal,
              'chunk_size': chunk_size}

    iterator = xmlutil.get_xml_iterator(source)

    return tree.VOTableFile(config=config).parse(iterator, config)


def parse_single_table(source, **kwargs):
    """
    Parses a VOTABLE xml file (or file-like object), assuming it only
    contains a single TABLE element, and returns a
    :class:`~vo.tree.Table` instance.

    See :func:`parse` for a description of the keyword arguments.
    """
    votable = parse(source, **kwargs)

    return votable.get_first_table()
