# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
This file defines the nodes that make up the VOTABLE XML tree.

.. _BINARY: http://www.ivoa.net/Documents/PR/VOTable/VOTable-20040322.html#ToC27
.. _COOSYS: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC19
.. _DESCRIPTION: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC19
.. _FIELD: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC24
.. _FIELDref: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC31
.. _GROUP: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC31
.. _ID: http://www.w3.org/TR/REC-xml/#id
.. _LINK: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC22
.. _multidimensional arrays: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC12
.. _numerical accuracy: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC26
.. _PARAM: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC24
.. _PARAMref: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC31
.. _RESOURCE: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC21
.. _TABLE: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC23
.. _TABLEDATA: http://www.ivoa.net/Documents/PR/VOTable/VOTable-20040322.html#ToC25
.. _unified content descriptor: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC28
.. _unique type: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC29
.. _units: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC27
.. _VALUES: http://www.ivoa.net/Documents/REC/VOTable/VOTable-20040811.html#ToC30
.. _VOTABLE: http://www.ivoa.net/Documents/PR/VOTable/VOTable-20040322.html#ToC9
"""

# STDLIB
import codecs
import cStringIO
import gzip
from math import ceil
from operator import attrgetter
import re
import urllib2

# THIRD-PARTY
import numpy as np
import pyfits

# LOCAL
import converters
from voexceptions import *
import util
import xmlutil

######################################################################
# FACTORY FUNCTIONS
def _lookup_by_id_factory(iterator, element_name, doc):
    """
    Creates a function useful for looking up an element by ID.

    *iterator* is a generator that iterates over some arbitrary set of
    elements,

    *element_name* should be the XML element name of the elements
    being iterated over (used for error messages only).

    *doc* is a docstring to apply to the generated function.
    """
    def lookup_by_id(self, ref, before=None):
        """
        Given an XML id *ref*, finds the first element in the iterator
        with the attribute ID == *ref*.  If *before* is provided, will
        stop searching at the object *before*.  This is important,
        since "forward references" are not allowed in the VOTABLE
        format.
        """
        for element in getattr(self, iterator)():
            if element is before:
                if element.ID == ref:
                    raise KeyError("%s references itself" % element_name)
                break
            if element.ID == ref:
                return element
        raise KeyError("No %s with ID '%s' found before the referencing %s" %
                       (element_name, ref, element_name))

    lookup_by_id.__doc__ = doc
    return lookup_by_id

######################################################################
# ATTRIBUTE CHECKERS
def check_id(ID, name='ID'):
    """
    Raises a :exc:`~vo.voexceptions.VOTableSpecError` if *ID* is not a
    valid XML ID_.  *name* is the name of the attribute being checked
    (used only for error messages).
    """
    if ID is not None and re.match(r"[A-Za-z_][A-Za-z0-9_\.\-:]*", ID) is None:
        raise VOTableSpecError("%s attribute '%s' is invalid" % (name, ID))

def check_ucd(ucd):
    """
    Raises a :exc:`~vo.voexceptions.VOTableSpecError` if *ucd* is not
    a valid `unified content descriptor`_ string as defined by the
    VOTABLE standard.
    """
    if ucd is not None and re.match(r"[A-Za-z0-9_\.;\\\-]*", ucd) is None:
        raise VOTableSpecError("ucd attribute '%s' is invalid" % ucd)

def check_astroyear(year):
    """
    Raises a :exc:`~vo.voexceptions.VOTableSpecError` if *year* is not
    a valid astronomical year as defined by the VOTABLE standard.
    """
    if year is not None and re.match(r"[JB]?[0-9]+([.][0-9]*)?", year) is None:
        raise VOTableSpecError("Invalid astroYear '%s'" % year)

def check_string(string, attr_name):
    """
    Raises a :exc:`~vo.voexceptions.VOTableSpecError` if *string* is
    not a string or Unicode string.
    """
    if type(string) not in (str, unicode):
        raise VOTableSpecError(
            "'%s' must be a string or Unicode string" % attr_name)


######################################################################
# ELEMENT CLASSES
class Element(object):
    """
    A base class for all classes that represent XML elements in the
    VOTABLE file.
    """
    pass

class _SimpleElement(Element):
    """
    A base class for simple elements, such as FIELD, PARAM and INFO
    that don't require any special parsing or outputting machinery.
    """
    def __init__(self):
        Element.__init__(self)

        self.description = None

    def parse(self, iterator, config):
        for start, tag, data in iterator:
            if start and tag not in ('DESCRIPTION', self._element_name):
                warn_or_raise("Unknown tag '%s'. Ignoring.")
            else:
                if tag == 'DESCRIPTION':
                    self.description = ''.join(data) or None
                elif tag == self._element_name:
                    break

        return self

    def to_xml(self, write, indent):
        indent_spaces = ' ' * indent
        write(indent_spaces)
        write('<%s ' % self._element_name)
        write(xmlutil.obj_to_attrs(self, self._attr_list))
        if self.description is not None:
            write('>\n')
            xmlutil.write_text_element(
                write, "DESCRIPTION", self.description, indent+1)
            write(indent_spaces)
            write('</%s>\n' % self._element_name)
        else:
            write('/>\n')


class Link(_SimpleElement):
    """
    A class for storing LINK_ elements, which are used to reference
    external documents and servers through a URI.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """
    _attr_list = ['ID', 'content_role', 'content_type', 'title', 'value',
                  'href', 'action']
    _element_name = 'LINK'

    def __init__(self, ID=None, title=None, value=None, href=None, action=None,
                 **kwargs):
        _SimpleElement.__init__(self)

        content_role = kwargs.get('content-role') or kwargs.get('content_role')
        content_type = kwargs.get('content-type') or kwargs.get('content_type')

        if 'gref' in kwargs:
            warn_or_raise(
                "The gref attribute on LINK is deprecated in VOTABLE 1.1")

        self.ID           = ID
        self.content_role = content_role
        self.content_type = content_type
        self.title        = title
        self.value        = value
        self.href         = href
        self.action       = action

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID_ of the LINK_ element.
        """)

    def _set_content_role(self, content_role):
        assert content_role in (None, 'query', 'hints', 'doc', 'location'), \
            "content-role attribute invalid."
        self._content_role = content_role
    def _del_content_role(self): self._content_role = None
    content_role = property(
        attrgetter('_content_role'), _set_content_role, _del_content_role,
        """
        Defines the MIME role of the referenced object.  Must be one of:

          None, 'query', 'hints', 'doc' or 'location'
        """)

    def _set_content_type(self, content_type):
        xmlutil.check_mime_content_type(content_type)
        self._content_type = content_type
    def _del_content_type(self): self._content_type = None
    content_type = property(
        attrgetter('_content_type'), _set_content_type, _del_content_type,
        """
        Defines the MIME content type of the referenced object.
        """)

    def _set_href(self, href):
        xmlutil.check_anyuri(href)
        self._href = href
    def _del_href(self): self._href = None
    href = property(
        attrgetter('_href'), _set_href, _del_href,
        """
        A URI to an arbitrary protocol.  The vo package only
        supports http and anonymous ftp.
        """)


class Info(_SimpleElement):
    """
    A class for storing INFO elements, which contain arbitrary
    key-value pairs for extensions to the standard.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """
    _attr_list = ['ID', 'name', 'value']
    _element_name = 'INFO'

    def __init__(self, ID=None, name=None, value=None):
        _SimpleElement.__init__(self)

        self.ID    = ID or name
        self.name  = name or ID
        self.value = value

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID_ of the INFO element.  Used for cross-referencing.
        """)

    def _set_name(self, name):
        assert name is not None, \
            "'name' attribute required for 'INFO' elements"
        xmlutil.check_token(name, 'name')
        self._name = name
    name = property(
        attrgetter('_name'), _set_name, None,
        """
        [*required*] The key of the key-value pair.
        """)

    def _set_value(self, value):
        assert value is not None, \
            "'value' attribute required for 'INFO' elements"
        check_string(value, 'value')
        self._value = value
    value = property(
        attrgetter('_value'), _set_value, None,
        """
        The value of the key-value pair.  (Always stored as a
        string or unicode string).
        """)


class Values(Element):
    """
    A class to represent the VALUES_ element, used within FIELD_ and
    PARAM_ elements to define the domain of values.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """
    def __init__(self, votable, field, ID=None, null=None, ref=None,
                 type="legal"):
        Element.__init__(self)

        self._votable = votable
        self._field   = field
        self.ID       = ID
        self.null     = null
        self._ref     = ref
        self.type     = type

        self.min           = None
        self.max           = None
        self.min_inclusive = True
        self.max_inclusive = True
        self._options       = []

    def _get_null(self): return self._field.converter.default
    def _set_null(self, null):
        if null is not None:
            self._field.converter.default = null
        else:
            self._field.converter.default = 0
    def _del_null(self): self._field.converter.default = 0
    null = property(
        _get_null, _set_null, _del_null,
        """
        For integral datatypes, *null* is used to define the value
        used for missing values.  Since the TABLEDATA_ datatype simply
        excludes missing values altogether, this attribute is ignored
        and merely stored for round-tripping support.  This attribute
        *should* be used for the BINARY_ format, but is not,
        currently.
        """)

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID of the VALUES_ element, used for cross-referencing.
        May be ``None`` or a string conforming to XML ID_ syntax.
        """)

    def _set_type(self, type):
        assert type in ('legal', 'actual'), 'type must be "legal" or "actual"'
        self._type = type
    type = property(
        attrgetter('_type'), _set_type, None,
        """
        [*required*] Defines the applicability of the domain defined
        by this VALUES_ element.  Must be one of the following
        strings:

          - 'legal': The domain of this column applies in general to
            this datatype. (default)

          - 'actual': The domain of this column applies only to the
            data enclosed in the parent table.
        """)

    def _set_ref(self, ref):
        check_id(ref, 'ref')
        if ref is not None:
            try:
                other = self._votable.get_values_by_id(ref, before=self)
            except KeyError:
                raise VOTableSpecError(
                    "VALUES ref='%s', which has not already been defined." %
                    self.ref)
            self.null = other.null
            self.type = other.type
            self.min = other.min
            self.min_inclusive = other.min_inclusive
            self.max = other.max
            self.max_inclusive = other.max_inclusive
            self.options = other.options
        self._ref = ref
    def _del_ref(self): self._ref = None
    ref = property(
        attrgetter('_ref'), _set_ref, _del_ref,
        """
        Refer to another VALUES_ element by ID_, defined previously in
        the document, for MIN/MAX/OPTION information.
        """)

    def _set_min(self, min):
        # TODO: datatype check
        self._min = min
    def _del_min(self): self._min = None
    min = property(
        attrgetter('_min'), _set_min, _del_min,
        """
        The minimum value of the domain.  See :attr:`min_inclusive`.
        """)

    def _set_min_inclusive(self, inclusive):
        if inclusive == 'yes':
            self._min_inclusive = True
        elif inclusive == 'no':
            self._min_inclusive = False
        else:
            self._min_inclusive = bool(inclusive)
    def _del_min_inclusive(self): self._min_inclusive = True
    min_inclusive = property(
        attrgetter('_min_inclusive'), _set_min_inclusive, _del_min_inclusive,
        """
        When True, the domain includes the minimum value.
        """)

    def _set_max(self, max):
        # TODO: datatype check
        self._max = max
    def _del_max(self): self._max = None
    max = property(
        attrgetter('_max'), _set_max, _del_max,
        """
        The maximum value of the domain.  See :attr:`max_inclusive`.
        """)

    def _set_max_inclusive(self, inclusive):
        if inclusive == 'yes':
            self._max_inclusive = True
        elif inclusive == 'no':
            self._max_inclusive = False
        else:
            self._max_inclusive = bool(inclusive)
    def _del_max_inclusive(self): self._max_inclusive = True
    max_inclusive = property(
        attrgetter('_max_inclusive'), _set_max_inclusive, _del_max_inclusive,
        """
        When True, the domain includes the maximum value.
        """)

    options = property(
        attrgetter('_options'), None, None,
        """
        A list of string key-value tuples defining other OPTION
        elements for the domain.  All options are ignored -- they are
        stored for round-tripping purposes only.
        """)

    def parse(self, iterator, config):
        if self.ref is not None:
            # To force the actual lookup of referenced element
            self.ref = self.ref
            for start, tag, data in iterator:
                if start:
                    raise VOTableSpecError(
                        "VALUES element with ref attribute has content.")
                else:
                    if tag != 'VALUES':
                        raise VOTableSpecError(
                            "VALUES element with ref attribute has content.")
                    break
        else:
            for start, tag, data in iterator:
                if start:
                    if tag == 'MIN':
                        assert 'value' in data, \
                            'MIN must have value attribute'
                        self.min = self._field.parser(data['value'])
                        self.min_inclusive = data.get('inclusive', 'yes')
                    elif tag == 'MAX':
                        assert 'value' in data, \
                            'MAX must have value attribute'
                        self.max = self._field.parser(data['value'])
                        self.max_inclusive = data.get('inclusive', 'yes')
                    elif tag == 'OPTION':
                        assert 'value' in data, \
                            'OPTION must have value attribute'
                        check_token(data.get('name'), 'name')
                        self.options.append(
                            (data.get('name'), data.get('value')))
                    elif tag != 'DESCRIPTION':
                        warn_or_raise("Unknown tag '%s'. Ignoring" % tag)
                else:
                    if tag == 'DESCRIPTION':
                        self.description = ''.join(data) or None
                    elif tag == 'VALUES':
                        break

        return self

    def to_xml(self, write, indent):
        def yes_no(value):
            if value: return 'yes'
            return 'no'

        indent_spaces = ' ' * indent

        if self.ref is not None:
            write(indent_spaces)
            write('<VALUES ' % self._element_name)
            write(xmlutil.obj_to_attrs(self, ['ref']))
            write('>\n')
        else:
            write(indent_spaces)
            write('<VALUES ' % self._element_name)
            write(xmlutil.obj_to_attrs(self, ['ID', 'null', 'ref']))
            write('>\n')
            if self.min is not None:
                write(indent_spaces)
                write(' <MIN value="%s" inclusive="%s"/>\n' %
                      (self._field.outputter(self.min,
                                             yes_no(self.min_inclusive))))
            if self.max is not None:
                write(indent_spaces)
                write(' <MAX value="%s" inclusive="%s"/>\n' %
                      (self._field.outputter(self.max,
                                             yes_no(self.max_inclusive))))
            for name, value in self.options:
                write(indent_spaces)
                write( '<OPTION ')
                if name is not None:
                    write('name="%s" ' % name)
                write('value="%s"/>\n' % value)
            write(indent_spaces)
            write('</VALUES>\n')


class Field(_SimpleElement):
    """
    A class that represents the FIELD_ element, which describes the
    datatype of a particular column of data.

    The keyword arguments correspond to setting members of the same
    name, documented below.

    If *ID* is provided, it is used for the column name in the
    resulting recarray of the table.  If no *ID* is provided, *name*
    is used instead.  If neither is provided, an exception will be
    raised.
    """
    _attr_list = ['ID', 'name', 'datatype', 'arraysize', 'ucd',
                  'unit', 'width', 'precision', 'utype', 'ref']
    _element_name = 'FIELD'

    def __init__(self, votable, ID=None, name=None, datatype=None,
                 arraysize=None, ucd=None, unit=None, width=None,
                 precision=None, utype=None, ref=None, type=None, config={}):
        _SimpleElement.__init__(self)

        self._votable = votable

        self.ID = ID or name
        self.name = name or ID
        if ID is None and name is None:
            raise VOTableSpecError(
                "<FIELD> element must have at least one of ID or name")

        if datatype == 'string':
            warn_or_raise(
                "'string' is not a valid VOTable datatype, treating as 'char'",
                config)
            datatype = 'char'
        elif datatype == 'unicodeString':
            warn_or_raise(
                "'unicodeString' is not a valid VOTable datatype, " +
                "treating as 'unicodeChar'",
                config)
            datatype = 'unicodeChar'

        self.ref       = ref
        self.datatype  = datatype
        self.arraysize = arraysize
        self.ucd       = ucd
        self.unit      = unit
        self.width     = width
        self.precision = precision
        self.utype     = utype
        self.values    = None
        self._links    = util.HomogeneousList(Link)

        self.title = self.name

    def _setup(self):
        self.converter = converters.get_converter(self)

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID of the FIELD_ element, used for cross-referencing.
        May be ``None`` or a string conforming to XML ID_ syntax.
        """)

    def _set_name(self, name):
        xmlutil.check_token(name, 'name')
        self._name = name
    def _del_name(self): self._name = None
    name = property(
        attrgetter('_name'), _set_name, _del_name,
        """
        An optional name for the FIELD_.
        """)

    def _set_datatype(self, datatype):
        assert datatype is not None, \
            "'datatype' attribute required for '%s' elements" % \
            self._element_name
        assert datatype in converters.converter_mapping, \
            "Invalid datatype attribute '%s'" % datatype
        self._datatype = datatype
    datatype = property(
        attrgetter('_datatype'), _set_datatype, None,
        """
        [*required*] The datatype of the column.  Valid values (as
        defined by the spec) are:

          'boolean', 'bit', 'unsignedByte', 'short', 'int', 'long',
          'char', 'unicodeChar', 'float', 'double', 'floatComplex', or
          'doubleComplex'

        Many VOTABLE files in the wild use 'string' instead of 'char',
        so that is also a valid option, though 'string' will always be
        converted to 'char' when writing the file back out.
        """)

    def _set_precision(self, precision):
        assert precision is None or re.match("[FE]?[0-9]+", precision), \
            "precision '%s' is invalid" % precision
        self._precision = precision
    def _del_precision(self): self._precision = None
    precision = property(
        attrgetter('_precision'), _set_precision, _del_precision,
        """
        Along with :attr:`width`, defines the `numerical accuracy`_
        associated with the data.  These values are used to limit the
        precision when writing floating point values back to the XML
        file.  Otherwise, it is purely informational -- the Numpy
        recarray containing the data itself does not use this
        information.
        """)

    def _set_width(self, width):
        if width is not None:
            width = int(width)
            assert width > 0, 'width attribute be a positive integer'
        self._width = width
    def _del_width(self): self._width = width
    width = property(
        attrgetter('_width'), _set_width, _del_width,
        """
        Along with :attr:`precision`, defines the `numerical
        accuracy`_ associated with the data.  These values are used to
        limit the precision when writing floating point values back to
        the XML file.  Otherwise, it is purely informational -- the
        Numpy recarray containing the data itself does not use this
        information.
        """)

    # ref on FIELD and PARAM behave differently than elsewhere -- here
    # they're just informational, such as to refer to a coordinate
    # system.
    def _set_ref(self, ref):
        check_id(ref, 'ref')
        self._ref = ref
    def _del_ref(self): self._ref = None
    ref = property(
        attrgetter('_ref'), _set_ref, _del_ref,
        """
        On FIELD_ elements, ref is used only for informational
        purposes, for example to refer to a COOSYS_ element.
        """
        )

    def _set_ucd(self, ucd):
        check_ucd(ucd)
        self._ucd = ucd
    def _del_ucd(self): self._ucd = None
    ucd = property(
        attrgetter('_ucd'), _set_ucd, _del_ucd,
        """
        The `unified content descriptor`_ for the FIELD_.
        """
        )

    def _set_utype(self, utype):
        self._utype = utype
    def _del_utype(self): self._utype = None
    utype = property(
        attrgetter('_utype'), _set_utype, _del_utype,
        """
        The usage-specific or `unique type`_ of the FIELD_.
        """
        )

    def _set_unit(self, unit):
        xmlutil.check_token(unit, 'unit')
        self._unit = unit
    def _del_unit(self): self._unit = None
    unit = property(
        attrgetter('_unit'), _set_unit, _del_unit,
        """
        A string specifying the units_ for the FIELD_.
        """)

    def _set_arraysize(self, arraysize):
        assert arraysize is None or \
            re.match("([0-9]+x)*[0-9]*[*]?(s\W)?", arraysize), \
            'arraysize attribute is invalid'
        self._arraysize = arraysize
    def _del_arraysize(self): self._arraysize = None
    arraysize = property(
        attrgetter('_arraysize'), _set_arraysize, _del_arraysize,
        """
        Specifies the size of the multidimensional array if this
        FIELD_ contains more than a single value.

        See `multidimensional arrays`_.
        """)

    def _set_type(self, type):
        if type is not None:
            warn("type attribute on FIELD is deprecated",
                 VOTableSpecWarning)
        self._type = type
    def _del_type(self): self._type = None
    type = property(
        attrgetter('_type'), _set_type, _del_type,
        """
        The type attribute on FIELD_ elements is deprecated, and not
        supported by the vo package.
        """)

    def _set_values(self, values):
        assert values is None or isinstance(values, Values)
        self._values = values
    def _del_values(self): self._values = None
    values = property(
        attrgetter('_values'), _set_values, _del_values,
        """
        A :class:`Values` instance (or ``None``) defining the domain of the
        column.
        """)

    links = property(
        attrgetter('_links'), None, None,
        """
        A list of :class:`Link` instances used to reference more details
        about the meaning of the FIELD_.  This is purely informational
        and is not used by the vo package.
        """)

    def parse(self, iterator, config):
        for start, tag, data in iterator:
            if start:
                if tag == 'VALUES':
                    if self.values is not None:
                        warn_or_raise(
                            "Too many values in field '%s'" % self.name,
                            config)
                    self.values = Values(
                        self._votable, self, **data).parse(iterator, config)
                elif tag == 'LINK':
                    self.links.append(Link(**data).parse(iterator, config))
                elif tag not in ('DESCRIPTION', self._element_name):
                    warn_or_raise("Unknown tag '%s'. Ignoring" % tag)
            else:
                if tag == 'DESCRIPTION':
                    if self.description is not None:
                        warn_or_raise(
                            "%s element contains more than one DESCRIPTION" %
                            self._element_name, config)
                    self.description = ''.join(data) or None
                elif tag == self._element_name:
                    break

        if self.description is not None:
            self.title = " ".join(x.strip() for x in
                                  self.description.split("\n"))
        else:
            self.title = self.name

        self._setup()

        return self

    def to_xml(self, write, indent):
        indent_spaces = ' ' * indent
        write(indent_spaces)
        write('<%s ' % self._element_name)
        write(xmlutil.obj_to_attrs(self, self._attr_list))
        if self.description is not None or self.values is not None:
            write('>\n')
            if self.description is not None:
                xmlutil.write_text_element(
                    write, "DESCRIPTION", self.description, indent+1)
            if self.values is not None:
                self.values.to_xml(write, indent+1)
            write(indent_spaces)
            write('</%s>\n' % self._element_name)
        else:
            write('/>\n')


class Param(Field):
    """
    A class to represent the PARAM_ element, which are constant-valued
    columns in the data.

    :class:`Param` objects are a subclass of :class:`Field`, and have
    all of its methods and members.  Additionally, it defines :attr:`value`.
    """
    _attr_list = Field._attr_list + ['value']
    _element_name = 'PARAM'

    def __init__(self, votable, ID=None, name=None, value=None, datatype=None,
                 arraysize=None, ucd=None, unit=None, width=None,
                 precision=None, utype=None, type=None, config={}):
        Field.__init__(self, votable, ID=ID, name=name, datatype=datatype,
                       arraysize=arraysize, ucd=ucd, unit=unit,
                       precision=precision, utype=utype, type=type,
                       config=config)
        self._value = value

    def _set_value(self, value):
        assert value is not None, 'value is required for all PARAM'
        if type(value) in (str, unicode):
            self._value = self.converter.parse(value)
        else:
            self._value = value
    value = property(
        attrgetter('_value'), _set_value, None,
        """
        [*required*] The constant value of the parameter.  Its type is
        determined by the :attr:`~Field.datatype` member.
        """)

    def _setup(self):
        Field._setup(self)
        self.value = self._value

    def to_xml(self, write, indent):
        tmp_value = self._value
        self._value = self.converter.output(tmp_value)
        Field.to_xml(self, write, indent)
        self._value = tmp_value


class CooSys(_SimpleElement):
    """
    A class representing the COOSYS_ element, which defines a
    coordinate system.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """
    _attr_list = ['ID', 'equinox', 'epoch', 'system']
    _element_name = 'COOSYS'

    def __init__(self, ID=None, equinox=None, epoch=None, system=None):
        _SimpleElement.__init__(self)

        self.ID      = ID
        self.equinox = equinox
        self.epoch   = epoch
        self.system  = system

    def _set_ID(self, ID):
        assert ID is not None, "ID is required for COOSYS"
        check_id(ID)
        self._ID = ID
    ID = property(
        attrgetter('_ID'), _set_ID, None,
        """
        [*required*] The XML ID of the COOSYS_ element, used for cross-referencing.
        May be ``None`` or a string conforming to XML ID_ syntax.
        """)

    def _set_system(self, system):
        assert system in ('eq_FK4', 'eq_FK5', 'ICRS', 'ecl_FK4', 'ecl_FK5',
                          'galactic', 'supergalactic', 'xy', 'barycentric',
                          'geo_app'), \
            "Invalid system attribute '%s'" % system
        self._system = system
    def _del_system(self): self._system = None
    system = property(
        attrgetter('_system'), _set_system, _del_system,
        """
        Specifies the type of coordinate system.  Valid choices are:

          'eq_FK4', 'eq_FK5', 'ICRS', 'ecl_FK4', 'ecl_FK5', 'galactic',
          'supergalactic', 'xy', 'barycentric', or 'geo_app'
        """)

    def _set_equinox(self, equinox):
        check_astroyear(equinox)
        self._equinox = equinox
    def _del_equinox(self): self._equinox = None
    equinox = property(
        attrgetter('_equinox'), _set_equinox, _del_equinox,
        """
        A parameter required to fix the equatorial or ecliptic systems
        (as e.g. "J2000" as the default "eq_FK5" or "B1950" as the
        default "eq_FK4").
        """)

    def _set_epoch(self, epoch):
        check_astroyear(epoch)
        self._epoch = epoch
    def _del_epoch(self): self._epoch = None
    epoch = property(
        attrgetter('_epoch'), _set_epoch, _del_epoch,
        """
        Specifies the epoch of the positions.  It must be a string
        specifying an astronomical year.
        """)


class FieldRef(_SimpleElement):
    """
    A class representing the FIELDref_ element, which is used inside
    of GROUP_ elements to refer to FIELD_ elements defined elsewhere.
    """
    _attr_list = ["ref"]
    _element_name = "FIELDref"

    def __init__(self, table, ref):
        """
        *table* is the :class:`Table` object that this :class:`FieldRef`
        is a member of.

        *ref* is the ID to reference a :class:`Field` object defined
         elsewhere.
        """
        _SimpleElement.__init__(self)
        self._table = table
        self.ref = ref

    def _set_ref(self, ref):
        check_id(ref, 'ref')
        self._ref = ref
    def _del_ref(self): self._ref = None
    ref = property(
        attrgetter('_ref'), _set_ref, _del_ref,
        """
        The ID_ of the FIELD_ that this FIELDref_ references.
        """)

    def get_ref(self):
        """
        Lookup the :class:`Field` instance that this :class:`FieldRef`
        references.
        """
        for field in self._table.iter_fields_and_params():
            if isinstance(field, Field) and field.ID == self.ref:
                return field
        raise ValueError("No field named '%s'" % self.ref)


class ParamRef(Element):
    """
    A class representing the PARAMref_ element, which is used inside
    of GROUP_ elements to refer to PARAM_ elements defined elsewhere.

    The keyword arguments correspond to setting members of the same
    name, documented below.

    It contains the following publicly-accessible members:

      *ref*: An XML ID refering to a <PARAM> element.
    """
    _attr_list = ["ref"]
    _element_name = "PARAMref"

    def __init__(self, table, ref):
        Element.__init__(self)
        self._table = table
        self.ref    = ref

    def _set_ref(self, ref):
        check_id(ref, 'ref')
        self._ref = ref
    def _del_ref(self): self._ref = None
    ref = property(
        attrgetter('_ref'), _set_ref, _del_ref,
        """
        The ID_ of the PARAM_ that this PARAMref_ references.
        """
        )

    def get_ref(self):
        """
        Lookup the :class:`Param` instance that this :class:`PARAMref`
        references.
        """
        for param in self._table.iter_fields_and_params():
            if isinstance(param, Param) and param.ID == self.ref:
                return param
        raise ValueError("No params named '%s'" % self.ref)


class Group(Element):
    """
    Stores information about the grouping of FIELD_ and PARAM_ elements.

    This information is currently ignored by the vo package---that is
    the columns in the recarray are always flat---but the grouping
    information is stored so that it can be written out again to the
    XML file.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """

    def __init__(self, table, ID=None, name=None, ref=None, ucd=None,
                 utype=None):
        Element.__init__(self)
        self._table = table

        self.ID          = ID or name
        self.name        = name or ID
        self.ref         = ref
        self.ucd         = ucd
        self.utype       = utype
        self.description = None

        self._entries = util.HomogeneousList((FieldRef, ParamRef, Group, Param))

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID of the GROUP_ element.  May be ``None`` or a string
        conforming to XML ID_ syntax.
        """)

    def _set_name(self, name):
        xmlutil.check_token(name, 'name')
        self._name = name
    def _del_name(self): self._name = None
    name = property(
        attrgetter('_name'), _set_name, _del_name,
        """
        An optional name for the grouping.
        """)

    def _set_ref(self, ref):
        check_id(ref, 'ref')
        self._ref = ref
    def _del_ref(self): self._ref = None
    ref = property(
        attrgetter('_ref'), _set_ref, _del_ref,
        """
        Currently ignored, as it's not clear from the spec
        how this is meant to work.
        """)

    def _set_ucd(self, ucd):
        check_ucd(ucd)
        self._ucd = ucd
    def _del_ucd(self): self._ucd = None
    ucd = property(
        attrgetter('_ucd'), _set_ucd, _del_ucd,
        """
        The `unified content descriptor`_ for the GROUP_.
        """)

    def _set_utype(self, utype):
        self._utype = utype
    def _del_utype(self): self._utype = None
    utype = property(
        attrgetter('_utype'), _set_utype, _del_utype,
        """
        The usage-specific or `unique type`_ of the GROUP_.
        """)

    def _set_description(self, description):
        self._description = description
    def _del_description(self): self._description = None
    description = property(
        attrgetter('_description'), _set_description, _del_description,
        """
        An optional string describing the GROUP_.  Corresponds to the
        DESCRIPTION_ element.
        """)

    entries = property(
        attrgetter('_entries'), None, None,
        """
        A list of members of the GROUP_.  This list may only contain
        objects of type :class:`Param`, :class:`Group`,
        :class:`ParamRef` and :class:`FieldRef`.
        """)

    def parse(self, iterator, config):
        for start, tag, data in iterator:
            if start:
                if tag == 'FIELDref':
                    self.entries.append(
                        FieldRef(self._table, **data).parse(iterator, config))
                elif tag == 'PARAMref':
                    self.entries.append(
                        ParamRef(self._table, **data).parse(iterator, config))
                elif tag == 'PARAM':
                    self.entries.append(
                        Param(self._votable, config=config, **data).parse(
                            iterator, config))
                elif tag == 'GROUP':
                    self.entries.append(
                        Group(self._table, **data).parse(iterator, config))
                elif tag != 'DESCRIPTION':
                    warn_or_raise("Unknown tag '%s'.  Ignoring." % tag)
            else:
                if tag == 'DESCRIPTION':
                    if self.description is not None:
                        warn_or_raise(
                            "GROUP element contains more than one DESCRIPTION",
                            config)
                    self.description = ''.join(data) or None
                elif tag == 'GROUP':
                    break
        return self

    def to_xml(self, write, indent):
        indent_spaces = ' ' * indent
        write(indent_spaces)
        write('<GROUP ')
        write(xmlutil.obj_to_attrs(self, ['ID', 'name', 'ref', 'ucd', 'utype']))
        write('>\n')
        if self.description is not None:
            xmlutil.write_text_element(
                write, "DESCRIPTION", self.description, indent+1)
        for entry in self.entries:
            entry.to_xml(write, indent+1)
        write(indent_spaces)
        write('</%s>\n' % self._element_name)

    def iter_fields_and_params(self):
        """
        Recursively iterate over all :class:`Param` elements in this
        :class:`Group`.
        """
        for entry in self.entries:
            if isinstance(entry, Param):
                yield entry
            elif isinstance(entry, Group):
                for field in entry.iter_fields_and_params():
                    yield field

    def iter_groups(self):
        """
        Recursively iterate over all sub-:class:`Group` instances in
        this :class:`Group`.
        """
        for entry in self.entries:
            if isinstance(entry, Group):
                yield entry
                for group in entry.iter_groups():
                    yield group


def uniqify_names(fields):
    """
    Make sure that all names and titles in a list of fields are
    unique, by appending numbers if necessary.
    """
    unique = {}
    for field in fields:
        if field.name in unique:
            raise ValueError("Duplicate field name '%s'" % field.name)
        unique[field.name] = None

    for field in fields:
        i = 0
        new_title = field.title
        while new_title in unique:
            new_title = field.title + " %02d" % i
            i += 1
        field.title = new_title
        unique[new_title] = None


class Table(Element):
    """
    A class to store a TABLE_ element, which optionally contains data.

    It contains the following publicly-accessible members, all of
    which are mutable:

      *array*: A Numpy recarray of the data itself, where each row is
      a row of votable data, and columns are named and typed based on
      the <FIELD> elements of the table.

      *mask*: A Numpy recarray of only boolean values, set to *True*
      wherever a value is undefined.

    If the Table contains no data, (for example, its enclosing
    :class:`Resource` has :attr:`~Resource.type` == 'meta') *array*
    and *mask* will be zero-length arrays.

    .. note::
      In a future version of the vo package, the *array* and *mask*
      elements will likely be combined into a single Numpy masked
      record array.  However, there are a number of deficiencies the
      current implementation of Numpy that prevent this.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """
    def __init__(self, votable, ID=None, name=None, ref=None, ucd=None,
                 utype=None, nrows=None):
        Element.__init__(self)
        self._votable = votable

        self.ID = ID or name
        self.name = name or ID
        check_id(ref)
        self._ref = ref
        self.ucd = ucd
        self.utype = utype
        if nrows is not None:
            nrows = int(nrows)
            assert nrows >= 0
        self._nrows = nrows
        self.description = None
        self.format = 'tabledata'

        self._fields = util.HomogeneousList(Field)
        self._params = util.HomogeneousList(Param)
        self._groups = util.HomogeneousList(Group)
        self._links  = util.HomogeneousList(Link)

        self.array = np.array([])
        self.mask  = np.array([])

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID of the TABLE_ element, used for cross-referencing.
        May be ``None`` or a string conforming to XML ID_ syntax.
        """)

    def _set_name(self, name):
        xmlutil.check_token(name, 'name')
        self._name = name
    def _del_name(self): self._name = None
    name = property(
        attrgetter('_name'), _set_name, _del_name,
        """
        An optional name for the table.
        """)

    def _set_ref(self, ref):
        # When the ref changes, we want to verify that it will work
        # by actually going and looking for the referenced table.
        # If found, set a bunch of properties in this table based
        # on the other one.
        check_id(ref, 'ref')
        if ref is not None:
            try:
                table = self._votable.get_table_by_id(ref, before=self)
            except KeyError:
                raise VOTableSpecError(
                    "Table ref='%s', which has not already been defined." %
                    self.ref)
            self.fields = table.fields
            self.params = table.params
            self.groups = table.groups
            self.links  = table.links
        self._ref = ref
    def _del_ref(self): self._ref = None
    ref = property(
        attrgetter('_ref'), _set_ref, _del_ref,
        """
        Refer to another TABLE, previously defined, by the *ref* ID_ for all
        metadata (FIELD_, PARAM_ etc.) information.
        """)

    def _set_ucd(self, ucd):
        check_ucd(ucd)
        self._ucd = ucd
    def _del_ucd(self): self._ucd = None
    ucd = property(
        attrgetter('_ucd'), _set_ucd, _del_ucd,
        """
        The `unified content descriptor`_ for the TABLE_.
        """
        )

    def _set_format(self, format):
        format = format.lower()
        if format == 'fits':
            raise ValueError("fits format can not be written out, only read.")
        assert format in ('tabledata', 'binary')
        self._format = format
    format = property(
        attrgetter('_format'), _set_format, None,
        """
        [*required*] The serialization format of the table.  Must be one of:

          'tabledata' (TABLEDATA_), 'binary' (BINARY_), 'fits' (FITS_).

        Note that the 'fits' format, since it requires an external
        file, can not be written out.  Any file read in with 'fits'
        format will be read out, by default, in 'tabledata' format.
        """)

    nrows = property(
        attrgetter('_nrows'), None, None,
        """
        [*immutable*] The number of rows in the table, as specified in
        the XML file.
        """)

    def _set_description(self, description):
        self._description = description
    def _del_description(self): self._description = None
    description = property(
        attrgetter('_description'), _set_description, _del_description,
        """
        An optional string describing the TABLE_.  Corresponds to the
        DESCRIPTION_ element.
        """)

    fields = property(
        attrgetter('_fields'), None, None,
        """
        A list of :class:`Field` objects describing the types of each
        of the data columns.
        """)

    params = property(
        attrgetter('_params'), None, None,
        """
        A list of parameters (constant-valued columns) for the
        table.  Must contain only :class:`Param` objects.
        """)

    groups = property(
        attrgetter('_groups'), None, None,
        """
        A list of :class:`Group` objects describing how the columns
        and parameters are grouped.  Currently this information is
        only kept around for round-tripping and informational
        purposes.
        """)

    links = property(
        attrgetter('_links'), None, None,
        """
        A list of :class:`Link` objects (pointers to other documents or servers
        through a URI) for the table.
        """)

    def create_arrays(self, array_data, mask_data, nrows=0):
        """
        Create new arrays to hold the data based on the current set of
        fields, and store them in the *array* and *mask* member
        variables.  Any data in existing arrays will be lost.

        *nrows*, if provided, is the number of rows to allocate.
        """
        if nrows is None:
            nrows = 0

        fields = self.fields
        uniqify_names(fields)
        for field in fields: field._setup()
        names = [x.name for x in fields]
        titles = [x.title for x in fields]
        formats = [x.converter.format for x in fields]

        descr = np.format_parser(formats, names, []).dtype
        array = np.array(array_data, dtype=descr)
        descr_mask = np.format_parser(['b1'] * len(fields), names, []).dtype
        mask = np.array(mask_data, dtype=descr_mask)

        self.array = array
        self.mask = mask

    def _resize_strategy(self, size):
        """
        Return a new (larger) size based on size, used for
        reallocating an array when it fills up.  This is in its own
        function to the resizing strategy can be easily replaced.
        """
        if size == 0:
            return 1
        return int(ceil(size * 1.5))

    def parse(self, iterator, config):
        columns = config['columns']

        if self.ref is not None:
            # This table doesn't have its own datatype descriptors, it
            # just references those from another table.

            # This is to call the property setter to go and get the
            # referenced information
            self.ref = self.ref

            for start, tag, data in iterator:
                if start:
                    if tag == 'DATA':
                        break
                else:
                    if tag == 'TABLE':
                        return self
                    elif tag == 'DESCRIPTION':
                        if self.description is not None:
                            warn_or_raise(
                                "RESOURCE contains more than one DESCRIPTION",
                                config)
                        self.description = ''.join(data) or None
        else:
            for start, tag, data in iterator:
                if start:
                    tag_mapping = {
                        'FIELD': lambda: self.fields.append(
                            Field(self._votable, config=config, **data).parse(
                                iterator, config)),
                        'PARAM': lambda: self.params.append(
                            Param(self._votable, config=config, **data).parse(
                                iterator, config)),
                        'GROUP': lambda: self.groups.append(
                            Group(self, **data).parse(iterator, config)),
                        'LINK': lambda: self.links.append(
                            Link(**data).parse(iterator, config)),
                        'DESCRIPTION': lambda: None
                        }

                    if tag == 'DATA':
                        break

                    tag_mapping.get(
                        tag,
                        lambda: warn_or_raise("Unknown tag '%s'. Ignoring" % tag))()
                else:
                    if tag == 'DESCRIPTION':
                        if self.description is not None:
                            warn_or_raise(
                                "RESOURCE contains more than one DESCRIPTION",
                                config)
                        self.description = ''.join(data) or None

        # self.create_arrays(self._nrows)
        fields = self.fields
        names = [x.name for x in fields]
        # Deal with a subset of the columns, if requested.
        if not columns:
            colnumbers = range(len(fields))
        else:
            if isinstance(columns, str):
                columns = [columns]
            try:
                colnumbers = [names.index(x) for x in columns]
            except ValueError:
                raise ValueError("Columns '%s' not found in fields list" %
                                 columns)

        for start, tag, data in iterator:
            if start:
                if tag == 'TABLEDATA':
                    self.array, self.mask = self._parse_tabledata(
                        iterator, colnumbers, fields, config)
                    break
                elif tag == 'BINARY':
                    self.array, self.mask = self._parse_binary(
                        iterator, colnumbers, fields, config)
                elif tag == 'FITS' :
                    try:
                        extnum = int(data.get('extnum', 0))
                        if extnum < 0:
                            raise ValueError()
                    except ValueError:
                        raise VOTableSpecError(
                            'extnum must be a positive integer')
                    self.array, self.mask = self._parse_fits(
                        iterator, extnum, config)
                else:
                    raise VOTableSpecError("Unknown data format %s" % tag)
        return self

    def _parse_tabledata(self, iterator, colnumbers, fields, config):
        # Since we don't know the number of rows up front, we'll
        # reallocate the record array to make room as we go.  This
        # prevents the need to scan through the XML twice.  The
        # allocation is by powers of 2.
        invalid = config['invalid']

#         array = self.array
#         mask = self.mask
#         # Need to have only one reference so that we can resize the
#         # array
#         del self.array
#         del self.mask

        array = []
        mask = []

        parsers = [field.converter.parse for field in fields]
        binparsers = [field.converter.binparse for field in fields]

        numrows = 0
#         alloc_rows = len(array)
        row_default = [x.converter.default for x in fields]
        mask_default = [True] * len(fields)
        rowdata = []
        for start, tag, data in iterator:
            if tag == 'TR':
                # Resize result arrays if necessary
#                 if numrows >= alloc_rows:
#                    alloc_rows = self._resize_strategy(alloc_rows)
#                    array.resize((alloc_rows,))
#                    mask.resize((alloc_rows,))

                # Parse the entire row into a list of strings
                del rowdata[:]
                binary = False
                for start, tag, data in iterator:
                   if start:
                        if tag == 'TD':
                            encoding = data.get('encoding', None) # EX
                            binary = (encoding == 'base64')       # EX
                        else:
                            warn_or_raise("Unknown tag '%s'. Ignoring." % tag)
                   else:
                        if tag == 'TD':
                            data = ''.join(data) or None
                            rowdata.append((data, binary))
                        elif tag == 'TR':
                            break
                        else:
                            warn_or_raise("Unknown tag '%s'. Ignoring." % tag)

                # Now convert only the requested columns to Python objects
                # and assign an entire row at a time to the recarray.
                row = row_default[:]
                row_mask = mask_default[:]
                for i in colnumbers:
                    value_str, binary = rowdata[i]
                    if value_str:
                        try:
                            if binary:
                                binparser = binparsers[i]
                                value = binparser(value_str.decode('base64'))
                            else:
                                parser = parsers[i]
                                value = parser(value_str)
                        except:
                            if invalid == 'exception':
                                raise
                        else:
                            row[i] = value
                            row_mask[i] = False

                array.append(tuple(row))
                mask.append(tuple(row_mask))
#                 array[numrows] = tuple(row)
#                 mask[numrows] = tuple(row_mask)
#                 numrows += 1
            elif tag == 'TABLE':
                break

        # Now, resize the array to the exact number of rows we need (it's
        # probably over-allocated).
        self.create_arrays(array, mask)

#         array = np.resize(array, (numrows,))
#         mask = np.resize(mask, (numrows,))

        if self.nrows >= 0 and self.nrows != len(array):
            warn('XML specified nrows="%d", but table contains %d rows' %
                 (self.nrows, numrows),
                 VOTableSpecWarning)
        self._nrows = len(array)

        return self.array, self.mask

    def _parse_binary(self, iterator, colnumbers, fields, config):
        fields = self.fields

        have_local_stream = False
        for start, tag, data in iterator:
            if tag == 'STREAM':
                if start:
                    if 'href' not in data:
                        have_local_stream = True
                        if data.get('encoding', None) != 'base64':
                            raise VOTableSpecError(
                                "Inline binary data must be base64 encoded")
                    else:
                        href = data['href']
                        encoding = data.get('encoding', None)
                else:
                    buffer = ''.join(data)
                    break

        if have_local_stream:
            buffer = buffer.decode('base64')
            string_io = cStringIO.StringIO(buffer)
            string_io.seek(0)
            read = string_io.read
        else:
            if not (href.startswith('http') or
                    href.startswith('ftp') or
                    href.startswith('file')):
                raise VOTableSpecError(
                    "The vo package only supports remote data through http, ftp or file")
            fd = urllib2.urlopen(href)
            if encoding is not None:
                if encoding == 'gzip':
                    fd = gzip.GzipFile(href, 'r', fileobj=fd)
                elif encoding == 'base64':
                    fd = codecs.EncodedFile(fd, 'base64')
                else:
                    raise VOTableSpecError(
                        "Unknown encoding type '%s'" % encoding)
            read = fd.read

        def careful_read(length):
            result = read(length)
            if len(result) != length:
                raise EOFError
            return result

        array = self.array
        mask = self.mask
        # Need to have only one reference so that we can resize the
        # array
        del self.array
        del self.mask

        binparsers = [field.converter.binparse for field in fields]

        numrows = 0
        alloc_rows = len(array)
        while True:
            # Resize result arrays if necessary
            if numrows >= alloc_rows:
                alloc_rows = self._resize_strategy(alloc_rows)
                array = np.resize(array, (alloc_rows,))
                mask = np.resize(mask, (alloc_rows,))

            row_data = []
            try:
                for binparse in binparsers:
                    row_data.append(binparse(careful_read))
            except EOFError:
                break

            row = [x.converter.default for x in fields]
            row_mask = [False] * len(fields)
            for i in colnumbers:
                row[i] = row_data[i]

            array[numrows] = tuple(row)
            mask[numrows] = tuple(row_mask)
            numrows += 1

        array = np.resize(array, (numrows,))
        mask = np.resize(mask, (numrows,))

        return array, mask

    def _parse_fits(self, iterator, extnum, config):
        fields = self.fields

        for start, tag, data in iterator:
            if tag == 'STREAM':
                if start:
                    href = data['href']
                    encoding = data.get('encoding', None)
                else:
                    break

        if not (href.startswith('http') or
                href.startswith('ftp') or
                href.startswith('file')):
            raise VOTableSpecError(
                "The vo package only supports remote data through http, ftp or file")

        fd = urllib2.urlopen(href)
        if encoding is not None:
            if encoding == 'gzip':
                fd = gzip.GzipFile(href, 'r', fileobj=fd)
            elif encoding == 'base64':
                fd = codecs.EncodedFile(fd, 'base64')
            else:
                raise VOTableSpecError(
                    "Unknown encoding type '%s'" % encoding)

        fits = pyfits.open(fd)

        array = fits[int(extnum)].data
        if array.dtype != self.array.dtype:
            warn_or_raise(
                "The fields defined in the VOTable do not match those in the " +
                "embedded FITS file.",
                config)

        return array, self.mask

    def to_xml(self, write, indent):
        indent_spaces = ' ' * indent
        write(indent_spaces)
        write('<TABLE ')
        write(xmlutil.obj_to_attrs(
                self,
                ('ID', 'name', 'ref', 'ucd', 'utype', 'nrows')))
        write('>\n')

        if self.description is not None:
            xmlutil.write_text_element(
                write, "DESCRIPTION", self.description, indent+1)

        for element_set in (self.fields, self.params):
            for element in element_set:
                element._setup()

        if self.ref is None:
            for element_set in (self.fields, self.params, self.groups,
                                self.links):
                for element in element_set:
                    element.to_xml(write, indent+1)

        if len(self.array):
            write(indent_spaces)
            write(' <DATA>\n')

            if self.format == 'fits':
                self.format = 'tabledata'

            if self.format == 'tabledata':
                self._write_tabledata(write, indent+2)
            elif self.format == 'binary':
                self._write_binary(write, indent+2)

            write(indent_spaces)
            write(' </DATA>\n')

        write(indent_spaces)
        write('</TABLE>\n')

    def _write_tabledata(self, write, indent):
        fields = self.fields
        array = self.array
        mask = self.mask

        indent_spaces = ' ' * indent
        tr_start = indent_spaces + " <TR>\n"
        tr_end = indent_spaces + " </TR>\n"
        td = indent_spaces + "  <TD>%s</TD>\n"
        td_empty = indent_spaces + "  <TD/>\n"
        fields = [(field.name, field.converter.output) for field in fields]

        write(indent_spaces)
        write('<TABLEDATA>\n')
        for row in xrange(len(array)):
            write(tr_start)
            for name, output in fields:
                masked = mask[name][row]
                if not masked:
                    write(td % output(array[name][row]))
                else:
                    write(td_empty)
            write(tr_end)
        write(indent_spaces)
        write('</TABLEDATA>\n')

    def _write_binary(self, write, indent):
        fields = self.fields
        array = self.array
        mask = self.mask
        indent_spaces = ' ' * indent

        write(indent_spaces)
        write('<BINARY>\n')
        write(indent_spaces)
        write(' <STREAM encoding="base64">\n')

        fields = [(field.name, field.converter.binoutput) for field in fields]

        data = cStringIO.StringIO()
        for row in xrange(len(array)):
            for name, converter in fields:
                data.write(converter(array[name][row]))

        write(data.getvalue().encode('base64'))
        write(indent_spaces)
        write(' </STREAM>\n')
        write(indent_spaces)
        write('</BINARY>\n')

    def iter_fields_and_params(self):
        """
        Recursively iterate over all FIELD and PARAM elements in the TABLE.
        """
        for param in self.params:
            yield param
        for field in self.fields:
            yield field
        for group in self.groups:
            for field in group.iter_fields_and_params():
                yield field

    get_field_by_id = _lookup_by_id_factory(
        'iter_fields_and_params', 'FIELD or PARAM',
        """
        Looks up a FIELD or PARAM element by the given ID.
        """)

    def iter_groups(self):
        """
        Recursively iterate over all GROUP elements in the TABLE.
        """
        for group in self.groups:
            yield group
            for g in group.iter_groups():
                yield g

    get_group_by_id = _lookup_by_id_factory(
        'iter_groups', 'GROUP',
        """
        Looks up a GROUP element by the given ID.  Used by the group's
        "ref" attribute
        """)


class Resource(Element):
    """
    A class to store the information in a RESOURCE_ element.  Each
    resource may contain zero-or-more TABLE_ elements and zero-or-more
    nested RESOURCE_ elements.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """
    def __init__(self, name=None, ID=None, utype=None, type='results',
                 **kwargs):
        Element.__init__(self)
        self.name              = name
        self.ID                = ID
        self.utype             = utype
        self.type              = type
        self._extra_attributes = kwargs
        self.description       = None

        self._coordinate_systems = util.HomogeneousList(CooSys)
        self._params             = util.HomogeneousList(Param)
        self._infos              = util.HomogeneousList(Info)
        self._links              = util.HomogeneousList(Link)
        self._tables             = util.HomogeneousList(Table)
        self._resources          = util.HomogeneousList(Resource)

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID of the RESOURCE_ element, used for
        cross-referencing.  May be ``None`` or a string conforming to
        XML ID_ syntax.
        """)

    def _set_name(self, name):
        xmlutil.check_token(name, 'name')
        self._name = name
    def _del_name(self): self._name = None
    name = property(
        attrgetter('_name'), _set_name, _del_name,
        """
        An optional name for the RESOURCE_.
        """)

    def _set_type(self, type):
        assert type in ('results', 'meta'), 'type must be "results" or "meta"'
        self._type = type
    type = property(
        attrgetter('_type'), _set_type, None,
        """
        [*required*] The type of the resource.  Must be either:

          - 'results': This resource contains actual result values (default)

          - 'meta': This resource contains only datatype descriptions
            (FIELD_ elements), but no actual data.
        """)

    def _set_utype(self, utype):
        self._utype = utype
    def _del_utype(self): self._utype = None
    utype = property(
        attrgetter('_utype'), _set_utype, _del_utype,
        """
        The usage-specific or `unique type`_ of the FIELD_.
        """
        )

    extra_attributes = property(
        attrgetter('_extra_attributes'), None, None,
        """
        A dictionary of string keys to string values
        containing any extra attributes of the RESOURCE_ element that
        are not defined in the specification.  (The specification
        explicitly allows for extra attributes here, but nowhere else.)
        """)

    def _set_description(self, description):
        self._description = description
    def _del_description(self): self._description = None
    description = property(
        attrgetter('_description'), _set_description, _del_description,
        """
        An optional string describing the RESOURCE_.  Corresponds to the
        DESCRIPTION_ element.
        """
        )

    coordinate_systems = property(
        attrgetter('_coordinate_systems'), None, None,
        """
        A list of coordinate system definitions (COOSYS_ elements) for
        the RESOURCE_.  Must contain only :class:`CooSys` objects.
        """)

    infos = property(
        attrgetter('_infos'), None, None,
        """
        A list of informational parameters (key-value pairs)
        for the resource.  Must only contain :class:`Info` objects.
        """)

    params = property(
        attrgetter('_params'), None, None,
        """
        A list of parameters (constant-valued columns) for the
        resource.  Must contain only :class:`Param` objects.
        """)

    links = property(
        attrgetter('_links'), None, None,
        """
        A list of links (pointers to other documents or servers
        through a URI) for the resource.  Must contain only :class:`Link`
        objects.
        """)

    tables = property(
        attrgetter('_tables'), None, None,
        """
        A list of tables in the resource.  Must contain only
        :class:`Table` objects.
        """)

    resources = property(
        attrgetter('_resources'), None, None,
        """
        A list of nested resources inside this resource.  Must contain
        only :class:`Resource` objects.
        """)

    def parse(self, votable, iterator, config):
        for start, tag, data in iterator:
            if start:
                tag_mapping = {
                    'TABLE': lambda: self.tables.append(
                        Table(votable, **data).parse(iterator, config)),
                    'INFO': lambda: self.infos.append(
                        Info(**data).parse(iterator, config)),
                    'PARAM': lambda: self.params.append(
                        Param(votable, config=config, **data).parse(
                            iterator, config)),
                    'COOSYS': lambda: self.coordinate_systems.append(
                        CooSys(**data).parse(iterator, config)),
                    'RESOURCE': lambda: self.resources.append(
                        Resource(**data).parse(votable, iterator, config)),
                    'LINK': lambda: self.links.append(
                        Link(**data).parse(iterator, config)),
                    'DESCRIPTION': lambda: None
                    }

                tag_mapping.get(
                    tag,
                    lambda: warn("Unknown tag '%s' found.  Ignoring." % tag))()
            elif tag == 'DESCRIPTION':
                if self.description is not None:
                    warn_or_raise(
                        "RESOURCE contains more than one DESCRIPTION",
                        config)
                self.description = ''.join(data) or None

        return self

    def to_xml(self, write, indent):
        indent_spaces = ' ' * indent
        write(indent_spaces)
        write('<RESOURCE ')
        write(xmlutil.obj_to_attrs(self, ('ID', 'type', 'utype')))
        write(' ')
        write(' '.join('%s=%s' for x in self.extra_attributes.items()))
        write('>\n')
        if self.description is not None:
            xmlutil.write_text_element(
                write, "DESCRIPTION", self.description, indent+1)
        for element_set in (self.coordinate_systems, self.params, self.infos,
                            self.links, self.tables, self.resources):
            for element in element_set:
                element.to_xml(write, indent+1)
        write(indent_spaces)
        write('</RESOURCE>\n')

    def iter_tables(self):
        """
        Recursively iterates over all tables in the resource and
        nested resources.
        """
        for table in self.tables:
            yield table
        for resource in self.resources:
            for table in resource.iter_tables():
                yield table

    def iter_fields_and_params(self):
        """
        Recursively iterates over all FIELD_ and PARAM_ elements in
        the resource, its tables and nested resources.
        """
        for param in self.params:
            yield param
        for table in self.tables:
            for param in table.iter_fields_and_params():
                yield param
        for resource in self.resources:
            for param in resource.iter_fields_and_params():
                yield param

    def iter_coosys(self):
        """
        Recursively iterates over all the COOSYS_ elements in the
        resource and nested resources.
        """
        for coosys in self.coordinate_systems():
            yield coosys
        for resource in self.resources:
            for coosys in resource.iter_coosys():
                yield coosys


class VOTableFile(Element):
    """
    A class to represent the top-level VOTABLE_ element.

    The keyword arguments correspond to setting members of the same
    name, documented below.
    """

    def __init__(self, ID=None):
        Element.__init__(self)
        self.ID                 = ID
        self.description        = None
        self._coordinate_systems = util.HomogeneousList(CooSys)
        self._params             = util.HomogeneousList(Param)
        self._infos              = util.HomogeneousList(Info)
        self._resources          = util.HomogeneousList(Resource)

    def _set_ID(self, ID):
        check_id(ID)
        self._ID = ID
    def _del_ID(self): self._ID = None
    ID = property(
        attrgetter('_ID'), _set_ID, _del_ID,
        """
        The XML ID of the VOTABLE_ element, used for cross-referencing.
        May be ``None`` or a string conforming to XML ID_ syntax.
        """)

    def _get_version(self):
        return '1.1'
    version = property(
        _get_version, None, None,
        """
        The version of the VOTable specification that the file uses.
        Always "1.1", since the vo package is designed to only output
        version 1.1 files.
        """)

    def _set_description(self, description):
        self._description = description
    def _del_description(self): self._description = None
    description = property(
        attrgetter('_description'), _set_description, _del_description,
        """
        An optional string describing the VOTABLE_.  Corresponds to the
        DESCRIPTION_ element.
        """)

    coordinate_systems = property(
        attrgetter('_coordinate_systems'), None, None,
        """
        A list of coordinate system descriptions for the file.  Must
        contain only :class:`CooSys` objects.
        """)

    params = property(
        attrgetter('_params'), None, None,
        """
        A list of parameters (constant-valued columns) that apply to the
        entire file.  Must contain only :class:`Param` objects.
        """)

    infos = property(
        attrgetter('_infos'), None, None,
        """
        A list of informational parameters (key-value pairs) for the
        entire file.  Must only contain :class:`Info` objects.
        """)

    resources = property(
        attrgetter('_resources'), None, None,
        """
        A list of resources, in the order they appear in the file.
        Must only contain :class:`Resource` objects.
        """)

    def parse(self, iterator, config):
        for start, tag, data in iterator:
            if start:
                if tag == 'VOTABLE':
                    if 'version' not in data:
                        warn_or_raise(
                            "No version number specified in file. " +
                            "Assuming %s" % self.version, config)
                    elif data['version'] != self.version:
                        warn("The vo package is written for VOTABLE " +
                             "version %s, but this file is %s" %
                             (self.version, data['version']),
                             VOTableSpecWarning)
                    break
                else:
                    raise ValueError("File does not appear to be a VOTABLE")

        for start, tag, data in iterator:
            if start:
                tag_mapping = {
                    'PARAM': lambda: self.params.append(
                        Param(self, config=config, **data).parse(
                            iterator, config)),
                    'RESOURCE': lambda: self.resources.append(
                        Resource(**data).parse(
                            self, iterator, config)),
                    'COOSYS': lambda: self.coordinate_systems.append(
                        CooSys(**data).parse(iterator, config)),
                    'INFO': lambda: self.infos.append(
                        Info(**data).parse(iterator, config)),
                    'DEFINITIONS': lambda: warn_or_raise(
                        "The 'DEFINITIONS' element is deprecated in VOTable " +
                        "1.1.  Ignoring.",
                        config),
                    'DESCRIPTION': lambda: None
                    }

                tag_mapping.get(
                    tag,
                    lambda: warn_or_raise(
                        "Unknown tag '%s' found.  Ignoring." % tag))()

            elif tag == 'DESCRIPTION':
                if self.description is not None:
                    warn_or_raise(
                        "VOTABLE contains more than one DESCRIPTION",
                        config)
                self.description = ''.join(data) or None
        return self

    def to_xml(self, fd):
        """
        Write to an XML file.

        *fd* may be a path to a file, or a Python file-like object.
        """
        write = util.convert_to_write_function(fd)

        write('<?xml version="1.0" encoding="utf-8"?>\n')
        write('<VOTABLE version="%s"\n' % self.version)
        write(' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\n')
        write(' xsi:noNamespaceSchemaLocation="http://www.ivoa.net/xml/VOTable/VOTable/v%s">\n' % self.version)
        if self.description is not None:
            xmlutil.write_text_element(
                write, "DESCRIPTION", self.description, 1)
        for element_set in (self.coordinate_systems, self.params,
                            self.infos, self.resources):
            for element in element_set:
                element.to_xml(write, 1)
        write('</VOTABLE>\n')

    def iter_tables(self):
        """
        Iterates over all tables in the VOTable file in a "flat" way,
        ignoring the nesting of resources etc.
        """
        for resource in self.resources:
            for table in resource.iter_tables():
                yield table

    def get_first_table(self):
        """
        Often, you know there is only one table in the file, and
        that's all you need.  This method returns that first table.
        """
        return self.iter_tables().next()

    get_table_by_id = _lookup_by_id_factory(
        'iter_tables', 'TABLE',
        """
        Looks up a TABLE element by the given ID.  Used by the table
        "ref" attribute.
        """)

    def iter_fields_and_params(self):
        """
        Recursively iterate over all FIELD and PARAM elements in the
        VOTABLE file.
        """
        for resources in self.resources:
            for field in resource.iter_fields_and_params():
                yield field

    get_field_by_id = _lookup_by_id_factory(
        'iter_fields_and_params', 'FIELD',
        """
        Looks up a FIELD element by the given ID.  Used by the field's
        "ref" attribute.
        """)

    def iter_values(self):
        """
        Recursively iterate over all VALUES_ elements in the VOTABLE
        file.
        """
        for field in self.iter_fields_and_params():
            yield field.value

    get_values_by_id = _lookup_by_id_factory(
        'iter_values', 'VALUES',
        """
        Looks up a VALUES element by the given ID.  Used by the values
        "ref" attribute.
        """)

    def iter_groups(self):
        """
        Recursively iterate over all GROUP elements in the VOTABLE
        file.
        """
        for table in self.iter_tables():
            for group in table.iter_groups():
                yield group

    get_group_by_id = _lookup_by_id_factory(
        'iter_groups', 'GROUP',
        """
        Looks up a GROUP element by the given ID.  Used by the group's
        "ref" attribute
        """)

    def iter_coosys(self):
        """
        Recursively iterate over all COOSYS elements in the VOTABLE
        file.
        """
        for coosys in self.coordinate_systems:
            yield coosys
        for resource in self.resources:
            for coosys in resource.iter_coosys():
                yield coosys

    get_coosys_by_id = _lookup_by_id_factory(
        'iter_coosys', 'COOSYS',
        """Looks up a COOSYS element by the given ID.""")

    def set_all_tables_format(self, format):
        """
        Set the output storage format of all tables in the file.
        """
        for table in self.iter_tables():
            table.format = format
