# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Various utilities and cookbook-like things.
"""

# STDLIB
import inspect
import cStringIO
import tokenize

class HomogeneousList(list):
    """
    A subclass of list that contains only elements of a given type or
    types.
    """
    def __init__(self, types, values=[]):
        """
        *types* is a sequence of acceptable types.

        *values* (optional) is an initial set of values.
        """
        self._types = types
        list.__init__(self, values)

    def _assert(self, x):
        assert isinstance(x, self._types), \
            ("homogeneous list must contain only objects of type '%s'" % \
                 (self._types))

    def __iadd__(self, other):
        for x in other:
            self._assert(x)
        return list.__iadd__(self, other)

    def __setitem__(self, x):
        self._assert(x)
        return list.__setitem__(self, x)

    def append(self, x):
        self._assert(x)
        return list.append(self, x)

    def insert(self, i, x):
        self._assert(x)
        return list.insert(self, i, x)


def convert_to_write_function(fd):
    """
    Returns a function suitable for streaming output.

    *fd* may be:

      - a function that writes to a stream, in which case it is
        returned verbatim.

      - a file path, in which case it is opened, and the :meth:`write`
        method on the file object is returned.

      - an object with a :meth:`write` method, in which case that
        method is returned.
    """
    if callable(fd):
        write = fd
    elif not hasattr(fd, 'write'):
        fd = open(fd, 'wb')
        write = fd.write
    else:
        write = fd.write
        assert callable(write)

    return write

def _parse(token, src):
    '''Parse text for use with safeeval. Code was taken from a mailing list posting
    example provided by Fredrik Lund. This is fairly brittle to white space and comments
    in its current form'''

    if token[1] == "{":
        out = {}
        token = src.next()
        while token[1] == '\n':
            token = src.next()
        while token[1] != "}":
            key = _parse(token, src)
            token = src.next()
            if token[1] != ":":
                raise SyntaxError("malformed dictionary")
            value = _parse(src.next(), src)
            out[key] = value
            token = src.next()
            if token[1] == ",":
                token = src.next()
            if token[1] == '\n':
                token = src.next()
        return out
    elif token[1] == "[":
        out = []
        token = src.next()
        while token[1] == '\n':
            token = src.next()        
        while token[1] != "]":
            out.append(_parse(token, src))
            token = src.next()
            if token[1] == ",":
                token = src.next()
            if token[1] == '\n':
                token = src.next()
        return out
    elif token[0] == tokenize.STRING:
        return token[1][1:-1].decode("string-escape")
    elif token[0] == tokenize.NUMBER:
        try:
            return int(token[1], 0)
        except ValueError:
            return float(token[1])
    else:
        raise SyntaxError("malformed expression")

def safeeval(source):
    '''parse a dictionary from a file or file-like object. Only strings and
    numbers are permitted as primitives and the only containers allowed are
    lists and dictionaries. This is to ensure that use of this file cannot
    be used as a security breach.
    
    >>> print safeeval("{'test':'123','hehe':['hooray',0x10]}")
    {'test': '123', 'hehe': ['hooray', 16]}
    '''
    src = cStringIO.StringIO(source).readline
    src = tokenize.generate_tokens(src)
    return _parse(src.next(), src)


