# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.

#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.

#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Various XML-related utilities
"""

# STDLIB
from xml.parsers import expat
import re
import textwrap

from voexceptions import warn_or_raise, warn

def xml_escape(data):
    """
    Escape &, <, >, ", ', in a string of data.
    """
    return data \
        .replace('&', '&amp;') \
        .replace('>', '&gt;') \
        .replace('<', '&lt;') \
        .replace('"', '&quot;') \
        .replace("'", '&apos;')


def fast_iterparse(fd, buffersize = 2 ** 10):
    """
    Based on :class:`cElementTree.iterparse`, but doesn't ever build a
    tree at all.  This makes things much faster and more memory
    efficient.

    The iterator returns 3-tuples (*event*, *tag*, *data*):

    - *event*: may be either 'start' (when an element begins) or 'end'
      (when an element is completed).

    - *tag*: The name of the element

    - *data*: Depends on the value of *event*:

      - if *event* == 'start', data is a dictionary of attributes

      - if *event* == 'end', data is a list of strings, that when
        joined is the text content of the element
    """
    close_at_end = False
    if not callable(fd):
        if not hasattr(fd, 'read'):
            close_at_end = True
            fd = open(fd, 'rb')
        read = fd.read
    else:
        read = fd

    queue = []
    text = []
    def start(name, attr):
        queue.append((True, name, attr))
        del text[:]

    def end(name):
        queue.append((False, name, text[:]))

    parser = expat.ParserCreate()
    parser.returns_unicode = False
    parser.specified_attributes = True
    parser.StartElementHandler = start
    parser.EndElementHandler = end
    parser.CharacterDataHandler = text.append
    Parse = parser.Parse

    data = read(buffersize)
    while data:
        Parse(data, False)
        for elem in queue:
            yield elem
        del queue[:]
        data = read(buffersize)

    if close_at_end:
        fd.close()
    Parse('', True)
    for elem in queue:
        yield elem

# fast_iterparse = iterparser.IterParser

def get_xml_iterator(source):
    """
    Returns an iterator over the elements of an XML file.  See
    :class:`fast_iterparse` for more information.
    """
    context = fast_iterparse(source)
    iterator = iter(context)
    return iterator


def obj_to_attrs(obj, attrs):
    """
    Converts an object with a bunch of attributes into an
    XML attribute list (as a string).

    *obj* is any Python object

    *attrs* is a sequence of attribute names to pull from the object

    If any of the attributes is ``None``, it will not appear in the
    XML.
    """
    d = []
    for attr in attrs:
        if getattr(obj, attr) is not None:
            d.append((attr.replace('_', '-'),
                      xml_escape(str(getattr(obj, attr)))))
    return u' '.join(u'%s="%s"' % x for x in d)


def write_text_element(write, element_name, content, indent):
    """
    Writes out an XML element, wrapping the text as necessary.

    *write*: a function that writes to a stream

    *element_name*: The tag name of the element

    *content*: the text of the element

    *indent*: The number of spaces to indent
    """
    indent_spaces = ' ' * indent
    write('%s<%s>\n' % (indent_spaces, element_name))
    indent_extra_spaces = ' ' * (indent + 1)
    content = ' '.join(x.strip() for x in xml_escape(content).split('\n'))
    filled = textwrap.fill(content,
                           initial_indent=indent_extra_spaces,
                           subsequent_indent=indent_extra_spaces)
    write(filled)
    write('\n')
    write('%s</%s>\n' % (indent_spaces, element_name))

def check_id(ID, name='ID', config={}):
    """
    Raises a :exc:`~vo.voexceptions.VOTableSpecError` if *ID* is not a
    valid XML ID_.  *name* is the name of the attribute being checked
    (used only for error messages).
    """
    if ID is not None and re.match(r"^[A-Za-z_][A-Za-z0-9_\.\-]*$", ID) is None:
        warn_or_raise(
            "%s attribute '%s' is invalid, must be a standard xml id" %
            (name, ID), config)
        return False
    return True


def fix_id(ID, config={}):
    """
    Given an arbitrary string, create one that can be used as an xml
    id.  This is rather simplistic at the moment, since it just
    replaces non-valid characters with underscores.
    """
    if ID is None:
        return None
    if re.match(r"^[A-Za-z_][A-Za-z0-9_\.\-]*$", ID):
        return ID
    if len(ID):
        if config.get('pedantic'):
            warn(
                "Implicitly generating an ID from a name that is not a valid ID '%s'" %
                ID)
        return (re.sub(r"[^A-Za-z_]", '_', ID[0]) +
                re.sub(r"[^A-Za-z0-9_\.\-]", "_", ID[1:]))
    return ''


_token_regex = r"(?![\r\l\t ])[^\r\l\t]*(?![\r\l\t ])"
def check_token(token, attr_name, config={}):
    """
    Raises a :exc:`ValueError` if *token* is not a valid XML token, as
    defined by XML Schema Part 2.
    """
    if (token is not None and
        not (token == '' or
             re.match(
                "[^\r\n\t ]?([^\r\n\t ]| [^\r\n\t ])*[^\r\n\t ]?$", token))):
        raise ValueError(
            "'%s' must be a valid token" % attr_name)
    return True


def check_mime_content_type(content_type, config={}):
    """
    Raises a :exc:`ValueError` if *content_type* is not a valid MIME
    content type (syntactically at least), as defined by RFC 2045.
    """
    ctrls = ''.join([chr(x) for x in xrange(0, 0x20)])
    token_regex = '[^()<>@,;:\\\"/[\]?= %s\x7f]+' % ctrls
    if (content_type is not None and
        re.match(
            r'(?P<type>%s)/(?P<subtype>%s)$' % (token_regex, token_regex),
            content_type) is None):
        warn_or_raise(
            "'content-type' must be a valid MIME content type", config)
        return False
    return True


def check_anyuri(uri, config={}):
    """
    Raises a :exc:`ValueError` if *uri* is not a valid URI as defined in RFC
    2396.
    """
    if (uri is not None and
        re.match(
            r"(([a-zA-Z][0-9a-zA-Z+\-\.]*:)?/{0,2}[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?(#[0-9a-zA-Z;/?:@&=+$\.\-_!~*'()%]+)?",
            uri) is None):
        warn_or_raise(
            "'%s' is not a valid URI" % uri, config)
        return False
    return True
